use std::{collections::HashMap, process, thread::sleep, time::Duration};

use crate::{errors::{create_error, print_error, ErrorSubType, ErrorType}, virtualmachine::{DecompiledOperation, Machine, VMMemory, VMMemoryNativeFunction, VMMemoryNull, VMMemoryString}};

fn get_string_from_vmmem(mem: &VMMemory) -> String {
    let mut out = String::new();
    match mem {
        VMMemory::String(s) => out.push_str(&s.value),
        VMMemory::Number(n) => out.push_str(&n.value.to_string()),
        VMMemory::Boolean(b) => out.push_str(if b.value { "true" } else { "false" }),
        VMMemory::Null(_) => out.push_str("null"),
        VMMemory::Table(tbl) => {
            out.push_str("{");
                for val in &tbl.values {
                out.push_str(&get_string_from_vmmem(&val.key.clone()));
                out.push_str("=");
                out.push_str(&get_string_from_vmmem(&val.value.clone()));
                out.push_str(", ");
            }
            out.pop();
            out.pop();
            out.push_str("}");
        },
        VMMemory::Function(func) => {
            out.push_str("|function: ");
            out.push_str(&func.id.to_string());
            out.push_str("|");
        },
        VMMemory::NativeFunction(_) => {
            out.push_str("|native function|");
        },
    }
    return out;
}
fn get_type_str(mem: &VMMemory) -> String {
    match mem {
        VMMemory::String(_) => "string".to_string(),
        VMMemory::Number(_) => "number".to_string(),
        VMMemory::Boolean(_) => "boolean".to_string(),
        VMMemory::Null(_) => "null".to_string(),
        VMMemory::Table(_) => "table".to_string(),
        VMMemory::Function(_) => "function".to_string(),
        VMMemory::NativeFunction(_) => "native_function".to_string(),
    }
}
fn expect(mem: &VMMemory, expected_type: &str) -> bool {
    let actual_type = get_type_str(&mem);
    if actual_type == expected_type {
        return true;
    } else {
        return false;
    }
}
fn error(msg: String, machine: &Machine, op: &DecompiledOperation) {
    let err = create_error(&msg, op.pos, ErrorType::MachineError, ErrorSubType::RuntimeError);
    let curr = machine.call_stack.len()-1;
    let func = machine.call_stack[curr].func;
    print_error(&err, &machine.ctx[func]);
    process::exit(1);
}

fn ugass(_machine: &mut Machine, _op: &DecompiledOperation, args: Vec<VMMemory>) -> VMMemory {
    let mut out = String::new();
    for arg in args {
        out.push_str(&get_string_from_vmmem(&arg));
        out.push_str("  ");
    }
    println!("{}", out.trim_end());
    return VMMemory::Null(VMMemoryNull { variable_id: 0 });
}

fn bimba(machine: &mut Machine, op: &DecompiledOperation, args: Vec<VMMemory>) -> VMMemory {
    if args.len() < 1 || !expect(&args[0], "number") {
        error(format!("#1 number expected got {}", get_type_str(&args[0])), machine, op);
    }
    match &args[0] {
        VMMemory::Number(n) => {
            sleep(Duration::from_millis(n.value as u64));
        },
        _ => {}
    }
    return VMMemory::Null(VMMemoryNull { variable_id: 0 });
}

fn csomor(machine: &mut Machine, op: &DecompiledOperation, args: Vec<VMMemory>) -> VMMemory {
    if args.len() < 1 || !expect(&args[0], "string") {
        error(format!("#1 string expected got {}", get_type_str(&args[0])), machine, op);
    }
    match &args[0] {
        VMMemory::String(s) => {
            error(s.value.clone(), machine, op);
        },
        _ => {}
    }
    return VMMemory::Null(VMMemoryNull { variable_id: 0 });
}

fn tarh(_machine: &mut Machine, _op: &DecompiledOperation, args: Vec<VMMemory>) -> VMMemory {
    if args.len() < 1 {
        return VMMemory::String(VMMemoryString { value: String::from("null"), variable_id: 0 });
    }
    return VMMemory::String(VMMemoryString { value: get_type_str(&args[0]), variable_id: 0 });
}

pub fn generate() -> HashMap<String, VMMemory> {
    let mut mem: HashMap<String, VMMemory> = HashMap::new();

    mem.insert(String::from("ugass"), VMMemory::NativeFunction(VMMemoryNativeFunction { func: ugass, variable_id: 0 }));
    mem.insert(String::from("bimba"), VMMemory::NativeFunction(VMMemoryNativeFunction { func: bimba, variable_id: 0 }));
    mem.insert(String::from("csömör"), VMMemory::NativeFunction(VMMemoryNativeFunction { func: csomor, variable_id: 0 }));
    mem.insert(String::from("tarh"), VMMemory::NativeFunction(VMMemoryNativeFunction { func: tarh, variable_id: 0 }));

    return mem;
}